﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.Rest.Net
{
    /// <summary>
    /// A class that can take a set of routes and generate code
    /// for a client that can be used to interact with them.
    /// </summary>
    public class CodeGenerator
    {
        /// <summary>
        /// Generates a CSharp client for a given set of routes.
        /// </summary>
        /// <param name="routes"></param>
        /// <returns></returns>
        public static string GenerateCSharpClient(IList<Route> routes)
        {
            var writer = new CodeWriter();

            writer.WriteLine("public class Client").WriteLine("{").Indent();
            writer.WriteLine("public string BaseUrl = null;");

            GenerateCSharpRouteClasses(routes, writer, out List<string> routeClasses);

            writer.WriteBreak();
            writer.WriteLine("public Client(string baseUrl)").WriteLine("{").Indent();
            writer.WriteLine("this.BaseUrl = baseUrl;");

            foreach (var @class in routeClasses)
                writer.WriteLine($"this.{@class} = new {@class}Functions(this);");

            writer.Outdent().WriteLine("}");

            writer.Outdent().WriteLine("}");

            return writer.ToString();
        }

        private static void GenerateCSharpRouteClasses(IList<Route> routes, CodeWriter writer, out List<string> routeClasses)
        {
            var groups = new Dictionary<string, List<Route>>();

            //Go through all the routes and group them.
            foreach (var route in routes)
            {
                //Get the method that this route is tied to.
                var methodInfo = route.GetTarget().GetMethodInfo();

                //See if this method defines the group for us.
                var routeGroup = methodInfo.GetCustomAttribute<RouteGroup>();

                //If there wasn't a route group on the method, see if there is one on the declaring type.
                if (routeGroup == null)
                    routeGroup = methodInfo.DeclaringType.GetCustomAttribute<RouteGroup>();

                //Group this route.
                string group = (routeGroup != null ? routeGroup.Name : methodInfo.DeclaringType.Name);

                if (groups.ContainsKey(group))
                    groups[group].Add(route);
                else
                    groups.Add(group, new List<Route>() { route });
            }

            //Generate classes for these groups
            foreach (var group in groups)
                GenerateCSharpRouteClass(group.Key, group.Value, writer);

            //Set the route classes
            routeClasses = new List<string>();
            routeClasses.AddRange(groups.Select(group => group.Key));
        }

        private static void GenerateCSharpRouteClass(string name, List<Route> routes, CodeWriter writer)
        {
            writer.WriteBreak();

            writer.WriteLine($"public {name}Functions {name};");

            writer.WriteBreak();

            writer.WriteLine($"public class {name}Functions").WriteLine("{").Indent();

            writer.WriteLine("public Client Client;");

            writer.WriteBreak();

            writer.WriteLine($"public {name}Functions(Client client)").WriteLine("{").Indent();

            writer.WriteLine("this.Client = client;");

            writer.Outdent().WriteLine("}");

            foreach (var route in routes)
                GenerateCSharpRouteFunction(route, writer);

            writer.Outdent().WriteLine("}");
        }

        private static void GenerateCSharpRouteFunction(Route route, CodeWriter writer)
        {
            writer.WriteBreak();

            var methodInfo = route.GetTarget().GetMethodInfo();

            writer.WriteLine($"public void {methodInfo.Name}()").WriteLine("{").Indent();

            writer.Outdent().WriteLine("}");
        }

        /// <summary>
        /// Generates a Javascript client for a given set of routes.
        /// </summary>
        /// <param name="routes"></param>
        /// <returns></returns>
        public static string GenerateJavascriptClient(IList<Route> routes)
        {
            return null;
        }
    }
}
